@@ -13,6 +13,7 @@
 #include "eog-hig-dialog.h"
 #include "eog-window.h"
 #include "session.h"
+#include "eog-config-keys.h"
 
 static void open_uri_list_cb (EogWindow *window, GList *uri_list, gpointer data); 
 static GtkWidget* create_new_window (void);
@@ -75,13 +76,17 @@ open_window (LoadContext *ctx)
 	GError *error = NULL;
 	gboolean new_window;
 	GList *it;
+	GConfClient *client;
 
 	g_return_val_if_fail (ctx->iid != NULL, FALSE);
 
-	/* FIXME: load gconf key */
-	new_window = TRUE; /* gconf_client_get_bool (priv->client, "/apps/eog/window/open_new_window", NULL); */
+	client = gconf_client_get_default ();
+
+	new_window = gconf_client_get_bool (client, EOG_CONF_WINDOW_OPEN_NEW_WINDOW, NULL);
 	new_window = new_window && ((ctx->window == NULL) || eog_window_has_contents (ctx->window));
 
+	g_object_unref (client);
+
 	if (ctx->single_windows) 
 	{
 		for (it = ctx->uri_list; it != NULL; it = it->next) {
